#!/usr/bin/env python

# Copyright 2015 - 2017:
#   The Royal Institution for the Advancement of Learning McGill University,
#   Centre National de la Recherche Scientifique,
#   University of Southern California,
#   Concordia University
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json
import os
import uuid


class Exporter():

    def __init__(self, descriptor):
        self.descriptor = descriptor

    def convert_type(self, boutiques_type, is_integer=False, is_list=False):
        if is_list:
            return "List"
        if boutiques_type == "Flag":
            return "Boolean"
        if boutiques_type == "Number":
            if is_integer:
                return "Int64"
            return "Double"
        return boutiques_type

    def convert_input_or_output(self, input_or_output, is_output):
        param = {}
        param['name'] = input_or_output.get('name')
        param['id'] = input_or_output.get('id')
        if is_output:
            param['type'] = 'File'
        else:
            param['type'] = self.convert_type(input_or_output.get('type'),
                                              input_or_output.get('integer'),
                                              input_or_output.get('list'))
        param['isOptional'] = input_or_output.get('optional') or False
        param['isReturnedValue'] = is_output
        if input_or_output.get('default-value'):
            param['defaultValue'] = input_or_output.get('default-value')
        if input_or_output.get('description'):
            param['description'] = input_or_output.get('description')
        return param

    def carmin(self, output_file):
        carmin_desc = {}
        with open(self.descriptor, 'r') as fhandle:
            descriptor = json.load(fhandle)

        carmin_desc['identifier'] = str(uuid.uuid4())
        carmin_desc['name'] = descriptor.get('name')
        carmin_desc['version'] = descriptor.get('tool-version')
        carmin_desc['description'] = descriptor.get('description')
        carmin_desc['canExecute'] = True
        carmin_desc['parameters'] = []
        for inp in descriptor.get('inputs'):
            carmin_desc['parameters'].append(
                                        self.convert_input_or_output(inp,
                                                                     False))
        for output in descriptor.get('output-files'):
            carmin_desc['parameters'].append(
                                        self.convert_input_or_output(output,
                                                                     True))
        carmin_desc['properties'] = {}
        carmin_desc['properties']['boutiques'] = True
        if descriptor.get('tags'):
            for prop in descriptor.get('tags').keys():
                carmin_desc['properties'][prop] = descriptor['tags'][prop]
        carmin_desc['errorCodesAndMessages'] = []
        for errors in descriptor.get('error-codes'):
            obj = {}
            obj['errorCode'] = errors['code']
            obj['errorMessage'] = errors['description']
            carmin_desc['errorCodesAndMessages'].append(obj)

        with open(output_file, 'w') as fhandle:
            fhandle.write(json.dumps(carmin_desc, indent=4))



from __future__ import absolute_import

from .localExec import LocalExecutor
from .invocationSchemaHandler import generateInvocationSchema
from .validator import validate_descriptor
from .bids import validate_bids
from .publisher import Publisher
from .evaluate import evaluateEngine
from .bosh import *

__all__ = ['localExec',
           'invocationSchemaHandler',
           'validator',
           'bids',
           'publisher',
           'evaluate',
           'bosh']



#!/usr/bin/env python

# Copyright 2015 - 2017:
#   The Royal Institution for the Advancement of Learning McGill University,
#   Centre National de la Recherche Scientifique,
#   University of Southern California,
#   Concordia University
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


def evaluateEngine(executor, query):
    try:
        # TODO: improve splitting to not fail in valid situations
        layers = query.split("/")[:-1] if "/" in query else [query]
        conditions = query.split("/")[-1] if "/" in query else []
        if len(conditions) > 0:
            if "," in conditions:
                conditions = conditions.split(",")
            else:
                conditions = [conditions]

        objs = executor.desc_dict
        for layer in layers:
            objs = objs[layer]

        query_result = {}
        for obj in objs:
            eligible = True
            for cond in conditions:
                lhs, rhs = cond.split("=")
                try:
                    rhs = float(rhs)
                except ValueError:
                    if rhs == "False":
                        rhs = False
                    elif rhs == "True":
                        rhs = True

                if obj.get(lhs) != rhs:
                    if not (obj.get(lhs) is None and rhs is False):
                        eligible = False

            if eligible:
                if "output-files" in layers:
                    query_result[obj["id"]] = executor.out_dict.get(obj["id"])
                elif "inputs" in layers:
                    # print(obj)
                    query_result[obj["id"]] = executor.in_dict.get(obj["id"])
                elif "groups" in layers:
                    query_result[obj["id"]] = {}
                    for mem in obj['members']:
                        query_result[obj["id"]][mem] = executor.in_dict.get(mem)

        return query_result

    except Exception:
        print("Invalid query ({0}). See --help.".format(query))
        return {}



#!/usr/bin/env python

# Copyright 2015 - 2017:
#   The Royal Institution for the Advancement of Learning McGill University,
#   Centre National de la Recherche Scientifique,
#   University of Southern California,
#   Concordia University
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import re
import simplejson
import os.path as op
from boutiques.validator import DescriptorValidationError


# BIDS validation module
def validate_bids(descriptor, valid=False):

    if not valid:
        msg = "Please provide a Boutiques descriptor that has been validated."
        raise DescriptorValidationError(msg)

    errors = []

    # TODO: verify not only that all fields/keys exist, their properties, too

    # Ensure the command-line conforms to the BIDS app spec
    msg_template = "   CLIError: command-line doesn't match template: {}"
    cltemp = "mkdir -p OUTPUT_DIR; (.*) BIDS_DIR OUTPUT_DIR ANALYSIS_LEVEL"\
             " PARTICIPANT_LABEL SESSION_LABEL[\\s]*(.*)"
    cmdline = descriptor["command-line"]
    if len(re.findall(cltemp, cmdline)) < 1:
        errors += [msg_template.format(cltemp)]

    # Verify IDs are present which link to the OUTPUT_DIR
    # key bot as File and String
    ftypes = set(["File", "String"])
    msg_template = "   OutError: \"{}\" types for outdir do not match \"{}\""
    outtypes = set([inp["type"]
                    for inp in descriptor["inputs"]
                    if inp["value-key"] == "OUTPUT_DIR"])
    if outtypes != ftypes:
        errors += [msg_template.format(", ".join(outtypes), ", ".join(ftypes))]

    # Verify that analysis levels is an enumerable with some
    # subset of "paricipant", "session", and "group"
    choices = ["session", "participant", "group"]
    msg_template = " LevelError: \"{}\" is not a valid analysis level"
    alevels = [inp["value-choices"]
               for inp in descriptor["inputs"]
               if inp["value-key"] == "ANALYSIS_LEVEL"][0]
    errors += [msg_template.format(lv)
               for lv in alevels
               if lv not in choices]

    # Verify there is only a single output defined (the directory)
    msg_template = "OutputError: 0 or multiple outputs defined"
    if len(descriptor["output-files"]) != 1:
        errors += [msg_template]
    else:
        # Verify that the output shows up as an output
        msg_template = "OutputError: OUTPUT_DIR is not represented as an output"
        if descriptor["output-files"][0]["path-template"] != "OUTPUT_DIR":
            errors += [msg_template]

    errors = None if errors == [] else errors
    if errors is None:
        print("BIDS validation OK")
    else:
        raise DescriptorValidationError("Invalid BIDS app descriptor:"
                                        "\n"+"\n".join(errors))



#!/usr/bin/env python

# A script for handling invocation schema generation and
# validation for Boutiques application descriptors
# Requires jsonschema 2.5

import json
import jsonschema
import os
import sys
import argparse
from functools import reduce
from jsonschema import ValidationError
from boutiques.validator import validate_descriptor


# An exception class specific to invocations
class InvocationValidationError(ValidationError):
    pass


# Generate an invocation schema from a Boutiques application descriptor
def generateInvocationSchema(toolDesc, oname=None, validateWrtMetaSchema=True):
    # Subclass dictionary to act more like a ruby hash map
    class RMap(dict):
        def __getitem__(self, key): return dict.get(self, key)
    s = RMap(toolDesc)
    descName, inputs, groups = s['name'], s['inputs'], s['groups'] or []
    oname = descName + ".invocationSchema" if oname is None else oname
    schema = {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "title": str(oname),
        "description": "Invocation schema for " + str(descName.lower()) + ".",
        "type": "object",
        "additionalProperties": False
    }
    if inputs is None:
        return schema

    # Handle type constraints for single inputs
    def addTypeConstraints(h, ind):
        i = RMap(ind)
        id, type, target = i['id'], i['type'], None

        def undertype():
            if i['integer']:
                return {'type': 'integer'}
            elif i['value-choices']:
                return {'enum': i['value-choices']}
            elif type == "File":
                return {'type': 'string'}
            elif type == "Flag":
                return {'type': 'boolean'}
            else:
                return {'type': type.lower()}
        if i['list']:
            h[id] = {'type': 'array', 'items': undertype()}
            if i["min-list-entries"]:
                h[id]["minItems"] = i["min-list-entries"]
            if i["max-list-entries"]:
                h[id]["maxItems"] = i["max-list-entries"]
            target = h[id]["items"]
        else:
            h[id] = undertype()
            target = h[id]
        if not i["minimum"] is None:
            target["minimum"] = i["minimum"]  # Python treats 0 as false
        if not i["maximum"] is None:
            target["maximum"] = i["maximum"]
        if i["exclusive-minimum"]:
            target["exclusiveMinimum"] = i["exclusive-minimum"]
        if i["exclusive-maximum"]:
            target["exclusiveMaximum"] = i["exclusive-maximum"]
        return h
    schema["properties"] = reduce(addTypeConstraints, inputs, {})
    # Required inputs
    reqInputs = [x['id'] for x in [x for x in inputs if not x.get('optional')]]
    if len(list(reqInputs)) > 0:
        schema["required"] = reqInputs

    # Helper functions (assumes properly formed descriptor)
    def byInd(id):
        return [i for i in inputs if id == i['id']][0]

    def isFlag(id):
        return (byInd(id)['type'] == 'Flag')
    # Handle one-is-required groups
    g1r = [g for g in groups if g.get("one-is-required")]

    def reqMember(m):  # Generate code for a one-is-required member
        if isFlag(m):
            return {"properties": {m: {"enum": [True]}}, "required": [m]}
        else:
            return {"required": [m]}
    if len(g1r) > 0:
        schema["allOf"] = [{"anyOf": list(map(reqMember, g['members']))}
                           for g in g1r]

    # Handle requires and disables-inputs/mutex constraints
    def handleDisablesRequires(h, inval):
        i, h = RMap(inval), RMap(h)
        id, reqs = i['id'], i['requires-inputs']
        disbs = i['disables-inputs'] or []
        # Mutex group members added to disablees list
        mutex_group_members = [g for g in groups if g.get('mutually-exclusive')
                               and (id in g['members'])]
        for g in mutex_group_members:
            for m in [m for m in g['members'] if not m == id]:
                disbs.append(m)
        # Handle requires-inputs (flags have to be true to satisfy it)
        if reqs:
            reqMap = {}
            for r in reqs:
                if isFlag(r):
                    reqMap[r] = {"enum": [True]}
            if isFlag(id):
                h[id] = {}
                h[id]['anyOf'] = [
                    {"properties": reqMap, "required": reqs},
                    {"properties": {id: {"enum": [False]}}}
                ]
            else:
                h[id] = {"required": reqs}
                h[id]['properties'] = reqMap
        # Handle disables-inputs (false flags do not violate
        # the disabled criteria)
        if len(disbs) > 0:
            if h[id] is None:
                h[id] = RMap({})

            def makeSingleDisablesMap(dh, d):  # Flags can be absent or false
                if isFlag(d):
                    dh[d] = {"oneOf": [{"not": {}}, {"enum": [False]}]}
                else:
                    dh[d] = {"not": {}}
                return dh
            disbMap = reduce(makeSingleDisablesMap, disbs, {})
            if isFlag(id):  # Flags do not disable their targets when false
                if h[id].get('anyOf') is None:
                    h[id]['anyOf'] = [
                        {"properties": disbMap},
                        {"properties": {id: {"enum": [False]}}}
                    ]
                # Need to add disables properties to requires ones,
                # if the latter exist
                else:
                    h[id]['anyOf'][0]['properties'].update(disbMap)
            else:
                if h[id].get('properties') is None:
                    h[id]['properties'] = RMap({})
                h[id]['properties'].update(disbMap)
        return h
    schema["dependencies"] = reduce(handleDisablesRequires, inputs, {})
    # Validate wrt meta-schema if desired
    if validateWrtMetaSchema:
        validateSchema(schema)
    # Return final schema
    return schema


# Validate data with respect to the invocation schema
def validateSchema(s, d=None, **kwargs):
        # Check schema wrt meta-schema
        try:
                jsonschema.Draft4Validator.check_schema(s)
        except jsonschema.SchemaError as se:
                errExit("Invocation schema is invalid.\n" +
                        str(se.message), False)
        # Check data instance against schema
        if d:
                try:
                        jsonschema.validate(d, s)
                except ValidationError as e:
                        raise InvocationValidationError(e)
                if kwargs.get("verbose"):
                        print("Invocation Schema validation OK")


# Script exit helper
def errExit(msg, parser, err_code=1):
        sys.stderr.write('Error: ' + msg + '\n')
        sys.exit(err_code)



#!/usr/bin/env python

# Copyright 2015 - 2017:
#   The Royal Institution for the Advancement of Learning McGill University,
#   Centre National de la Recherche Scientifique,
#   University of Southern California,
#   Concordia University
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from argparse import ArgumentParser
from jsonschema import ValidationError
from boutiques.validator import validate_descriptor
import json
import os


class Importer():

    def __init__(self, output_file):
        self.output_file = output_file

    def upgrade_04(self, input_file):
        """
         Differences between 0.4 and current (0.5):
           -schema version (obv)
           -singularity should now be represented same as docker
           -walltime should be part of suggested_resources structure

        I.e.
        "schema-version": "0.4",
                    ...... becomes.....
        "schema-version": "0.5",

        I.e.
        "container-image": {
          "type": "singularity",
          "url": "shub://gkiar/ndmg-cbrain:master"
          },
                    ...... becomes.....
        "container-image": {
          "type": "singularity",
          "image": "gkiar/ndmg-cbrain:master",
          "index": "shub://",
        },

        I.e.
        "walltime-estimate": 3600,
                    ...... becomes.....
        "suggested-resources": {
          "walltime-estimate": 3600
        },
        """
        with open(input_file, 'r') as fhandle:
            descriptor = json.load(fhandle)

        descriptor["schema-version"] = "0.5"

        if "container-image" in descriptor.keys():
            if "singularity" == descriptor["container-image"]["type"]:
                url = descriptor["container-image"]["url"]
                img = url.split("://")
                if len(img) == 1:
                    descriptor["container-image"]["image"] = img[0]
                elif len(img) == 2:
                    descriptor["container-image"]["image"] = img[1]
                    descriptor["container-image"]["index"] = img[0] + "://"
                del descriptor["container-image"]["url"]
            elif ("docker" == descriptor["container-image"]["type"] and
                  descriptor["container-image"].get("index")):
                url = descriptor["container-image"]["index"].split("://")[-1]
                descriptor["container-image"]["index"] = url

        if "walltime-estimate" in descriptor.keys():
            descriptor["suggested-resources"] =\
              {"walltime-estimate": descriptor["walltime-estimate"]}
            del descriptor["walltime-estimate"]

        with open(self.output_file, 'w') as fhandle:
            fhandle.write(json.dumps(descriptor, indent=4))
        validate_descriptor(self.output_file)

    def get_entry_point(self, app_dir):
        entrypoint = None
        with open(os.path.join(app_dir, "Dockerfile")) as f:
            content = f.readlines()
        for line in content:
            split = line.split()
            if len(split) >= 2 and split[0] == "ENTRYPOINT":
                entrypoint = split[1].strip("[]\"")
        return entrypoint

    def import_bids(self, app_dir):
        path, fil = os.path.split(__file__)
        template_file = os.path.join(path, "bids-app-template", "template.json")

        with open(template_file) as f:
            template_string = f.read()

        errors = []
        app_name = os.path.basename(os.path.abspath(app_dir))
        with open(os.path.join(app_dir, "version"), "r") as f:
            version = f.read().strip()
        git_repo = "https://github.com/BIDS-Apps/"+app_name
        entrypoint = self.get_entry_point(app_dir)
        container_image = "bids/"+app_name
        analysis_types = "participant\", \"group\", \"session"

        if not entrypoint:
            errors.append("No entrypoint found in container.")

        if len(errors):
            raise ValidationError("Invalid descriptor:\n"+"\n".join(errors))

        template_string = template_string.replace("@@APP_NAME@@", app_name)
        template_string = template_string.replace("@@VERSION@@", version)
        template_string = template_string.replace("@@GIT_REPO_URL@@", git_repo)
        template_string = template_string.replace("@@DOCKER_ENTRYPOINT@@",
                                                  entrypoint)
        template_string = template_string.replace("@@CONTAINER_IMAGE@@",
                                                  container_image)
        template_string = template_string.replace("@@ANALYSIS_TYPES@@",
                                                  analysis_types)

        with open(self.output_file, "w") as f:
            f.write(template_string)



#!/usr/bin/env python

# Copyright 2015 - 2017:
#   The Royal Institution for the Advancement of Learning McGill University,
#   Centre National de la Recherche Scientifique,
#   University of Southern California,
#   Concordia University
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from boutiques.validator import validate_descriptor
import json
import requests
import os


class ZenodoError(Exception):
    pass


class Publisher():

    def __init__(self, descriptor_file_name,
                 creator, affiliation, verbose, sandbox, no_int,
                 auth_token):
        # Straightforward assignments
        self.verbose = verbose
        self.sandbox = sandbox
        self.descriptor_file_name = descriptor_file_name
        self.creator = creator
        self.affiliation = affiliation
        self.no_int = no_int
        self.zenodo_access_token = auth_token

        # Validate and load descriptor
        validate_descriptor(descriptor_file_name)
        self.descriptor = json.loads(open(self.descriptor_file_name).read())

        self.config_file = os.path.join(os.getenv("HOME"), ".boutiques")
        # Fix Zenodo access token
        if self.zenodo_access_token is None:
            self.zenodo_access_token = self.get_zenodo_access_token()
        self.save_zenodo_access_token()

        # Set Zenodo endpoint
        self.zenodo_endpoint = "https://sandbox.zenodo.org" if\
            self.sandbox else "https://zenodo.org"
        if(self.verbose):
            print("[ INFO ] Using Zenodo endpoint {0}".
                  format(self.zenodo_endpoint))

    def get_zenodo_access_token(self):
        json_creds = self.read_credentials()
        if json_creds.get(self.config_token_property_name()):
            return json_creds.get(self.config_token_property_name())
        if(self.no_int):
            raise ZenodoError("Cannot find Zenodo credentials.")
        prompt = ("Please enter your Zenodo access token (it will be "
                  "saved in {0} for future use): ".format(self.config_file))
        try:
            return raw_input(prompt)  # Python 2
        except NameError:
            return input(prompt)  # Python 3

    def save_zenodo_access_token(self):
        json_creds = self.read_credentials()
        json_creds[self.config_token_property_name()] = self.zenodo_access_token
        with open(self.config_file, 'w') as f:
            f.write(json.dumps(json_creds, indent=4, sort_keys=True))
        if(self.verbose):
            print("[ INFO ] Zenodo access token saved in {0}".
                  format(self.config_file))

    def config_token_property_name(self):
        if self.sandbox:
            return "zenodo-access-token-test"
        return "zenodo-access-token"

    def read_credentials(self):
        try:
            with open(self.config_file, "r") as f:
                json_creds = json.load(f)
        except IOError:
            json_creds = {}
        except ValueError:
            json_creds = {}
        return json_creds

    def zenodo_test_api(self):
        r = requests.get(self.zenodo_endpoint+'/api/deposit/depositions')
        if(r.status_code != 401):
            self.raise_zenodo_error("Cannot access Zenodo", r)
        if(self.verbose):
            self.print_zenodo_info("Zenodo is accessible", r)
        r = requests.get(self.zenodo_endpoint+'/api/deposit/depositions',
                         params={'access_token': self.zenodo_access_token})
        message = "Cannot authenticate to Zenodo API, check you access token"
        if(r.status_code != 200):
            self.raise_zenodo_error("Cannot authenticate to Zenodo", r)
        if(self.verbose):
            self.print_zenodo_info("Authentication to Zenodo successful", r)

    def zenodo_deposit(self):
        headers = {"Content-Type": "application/json"}
        data = {
            'metadata': {
                'title': self.descriptor['name'],
                'upload_type': 'software',
                'description': self.descriptor['description'] or "Boutiques "
                               "descriptor for {0}".format(
                                                   self.descriptor['name']),
                'creators': [{'name': self.creator,
                              'affiliation': self.affiliation}],
                'version': self.descriptor['tool-version'],
                'keywords': ['Boutiques',
                             'schema-version:{0}'.
                             format(self.descriptor['schema-version'])]
            }
        }
        keywords = data['metadata']['keywords']
        for tag in self.descriptor.get('tags'):
            keywords.append(tag + ":" + self.descriptor['tags'][tag])
        if self.descriptor.get('container-image'):
            keywords.append(self.descriptor['container-image']['type'])

        r = requests.post(self.zenodo_endpoint+'/api/deposit/depositions',
                          params={'access_token': self.zenodo_access_token},
                          json={},
                          data=json.dumps(data),
                          headers=headers)
        if(r.status_code != 201):
            self.raise_zenodo_error("Deposition failed", r)
        zid = r.json()['id']
        if(self.verbose):
            self.print_zenodo_info("Deposition succeeded, id is {0}".
                                   format(zid), r)
        return zid

    def zenodo_upload_descriptor(self, deposition_id):
        data = {'filename': os.path.basename(self.descriptor_file_name)}
        files = {'file': open(self.descriptor_file_name, 'rb')}
        r = requests.post(self.zenodo_endpoint +
                          '/api/deposit/depositions/%s/files'
                          % deposition_id,
                          params={'access_token': self.zenodo_access_token},
                          data=data,
                          files=files)
        # Status code is inconsistent with Zenodo documentation

        if(r.status_code != 201):
            self.raise_zenodo_error("Cannot upload descriptor", r)
        if(self.verbose):
            self.print_zenodo_info("Descriptor uploaded to Zenodo", r)

    def zenodo_publish(self, deposition_id):
        r = requests.post(self.zenodo_endpoint +
                          '/api/deposit/depositions/%s/actions/publish'
                          % deposition_id,
                          params={'access_token': self.zenodo_access_token})
        if(r.status_code != 202):
            self.raise_zenodo_error("Cannot publish descriptor", r)
        if(self.verbose):
            self.print_zenodo_info("Descriptor published to Zenodo, doi is {0}"
                                   .format(r.json()['doi']), r)

    def raise_zenodo_error(self, message, r):
        raise ZenodoError("Zenodo error ({0}): {1}."
                          .format(r.status_code, message))

    def print_zenodo_info(self, message, r):
        print("[ INFO ({1}) ] {0}".format(message, r.status_code))

    def publish(self):
        if(not self.no_int):
            prompt = ("The descriptor will be published to Zenodo, "
                      "this cannot be undone. Are you sure? (Y/n) ")
            try:
                ret = raw_input(prompt)  # Python 2
            except NameError:
                ret = input(prompt)  # Python 3
            if ret != "Y":
                return
        self.zenodo_test_api()
        deposition_id = self.zenodo_deposit()
        self.zenodo_upload_descriptor(deposition_id)
        self.zenodo_publish(deposition_id)



#!/usr/bin/env python

import jsonschema
import json
import os
import sys
import os.path as op
import tempfile
import pytest
from argparse import ArgumentParser, RawTextHelpFormatter
from jsonschema import ValidationError
from boutiques.validator import DescriptorValidationError
from boutiques.publisher import ZenodoError
from boutiques.invocationSchemaHandler import InvocationValidationError
from boutiques.localExec import ToolOutputNotFoundError


def validate(*params):
    parser = ArgumentParser("Boutiques descriptor validator")
    parser.add_argument("descriptor", action="store",
                        help="The Boutiques descriptor.")
    parser.add_argument("--bids", "-b", action="store_true",
                        help="Flag indicating if descriptor is a BIDS app")
    results = parser.parse_args(params)

    from boutiques.validator import validate_descriptor
    descriptor = validate_descriptor(results.descriptor)
    if results.bids:
        from boutiques.bids import validate_bids
        validate_bids(descriptor, valid=True)
    return 0


def execute(*params):
    parser = ArgumentParser("Boutiques local executor", add_help=False)
    parser.add_argument("mode", action="store",
                        help="Mode of operation to use. Launch: takes a "
                        "set of inputs compliant with invocation schema "
                        "and launches the tool. Simulate: shows sample "
                        "command-lines based on the provided descriptor"
                        " based on provided or randomly generated "
                        "inputs.", choices=["launch", "simulate"])
    parser.add_argument("--help", "-h", action="store_true",
                        help="show this help message and exit")

    helps = any([True for ht in ["--help", "-h"] if ht in params])
    if len(params) <= 1 and helps:
        parser.print_help()
        raise SystemExit

    args, params = parser.parse_known_args(params)
    mode = args.mode
    params += ["--help"] if args.help is True else []

    if mode == "launch":
        parser = ArgumentParser("Launches an invocation.")
        parser.add_argument("descriptor", action="store",
                            help="The Boutiques descriptor.")
        parser.add_argument("invocation", action="store",
                            help="Input JSON complying to invocation.")
        parser.add_argument("-v", "--volumes", action="store", type=str,
                            help="Volumes to mount when launching the "
                            "container. Format consistently the following:"
                            " /a:/b will mount local direcotry /a to "
                            "container directory /b.", nargs="*")
        parser.add_argument("-x", "--debug", action="store_true",
                            help="Keeps temporary scripts used during "
                            "execution.")
        parser.add_argument("-u", "--user", action="store_true",
                            help="Runs the container as local user ({0})"
                            " instead of root.".format(os.getenv("USER")))
        results = parser.parse_args(params)
        descriptor = results.descriptor
        inp = results.invocation

        # Do some basic input scrubbing
        if not os.path.isfile(inp):
            raise SystemExit("Input file {} does not exist".format(inp))
        if not inp.endswith(".json"):
            raise SystemExit("Input file {} must end in json".format(inp))
        if not os.path.isfile(descriptor):
            raise SystemExit("JSON descriptor {} does not exist".
                             format(descriptor))

        # Validate invocation and descriptor
        valid = invocation(descriptor, '-i', inp)

        # Generate object that will perform the commands
        from boutiques.localExec import LocalExecutor
        executor = LocalExecutor(descriptor,
                                 {"forcePathType": True,
                                  "destroyTempScripts": not results.debug,
                                  "changeUser": results.user})
        executor.readInput(inp)
        # Execute it
        stdout, stderr, exit_code, err_msg = executor.execute(results.volumes)
        return stdout, stderr, exit_code, err_msg

    if mode == "simulate":
        parser = ArgumentParser("Simulates an invocation.")
        parser.add_argument("descriptor", action="store",
                            help="The Boutiques descriptor.")
        parser.add_argument("-i", "--input", action="store",
                            help="Input JSON complying to invocation.")
        parser.add_argument("-r", "--random", action="store", type=int,
                            nargs="*", help="Generate random set of inputs.")
        results = parser.parse_args(params)
        descriptor = results.descriptor

        # Do some basic input scrubbing
        inp = results.input
        rand = results.random is not None
        numb = results.random[0] if rand and len(results.random) > 0 else 1

        if numb and numb < 1:
            raise SystemExit("--number value must be positive.")
        if rand and inp:
            raise SystemExit("--random setting and --input value cannot "
                             "be used together.")
        if inp and not os.path.isfile(inp):
            raise SystemExit("Input file {} does not exist.".format(inp))
        if inp and not inp.endswith(".json"):
            raise SystemExit("Input file {} must end in 'json'.".format(inp))
        if not os.path.isfile(descriptor):
            raise SystemExit("JSON descriptor {} does not seem to exist."
                             .format(descriptor))
        if not rand and not inp:
            raise SystemExit("The default mode requires an input (-i).")

        valid = invocation(descriptor, '-i', inp) if inp else\
            invocation(descriptor)

        # Generate object that will perform the commands
        from boutiques.localExec import LocalExecutor
        executor = LocalExecutor(descriptor,
                                 {"forcePathType": True,
                                  "destroyTempScripts": True,
                                  "changeUser": True})
        if rand:
            executor.generateRandomParams(numb)
            executor.printCmdLine()
        else:
            executor.readInput(inp)
            executor.printCmdLine()

        return "", "", 0, ""  # for consistency with execute


def importer(*params):
    parser = ArgumentParser("Imports old descriptor or BIDS app to spec.")
    parser.add_argument("type", help="Type of import we are performing",
                        choices=["bids", "0.4"])
    parser.add_argument("descriptor", help="Where the Boutiques"
                        " descriptor will be written.")
    parser.add_argument("input", help="Input to be convered. For '0.4'"
                        ", is JSON descriptor,"
                        " for 'bids' is base directory of BIDS app.")
    results = parser.parse_args(params)

    descriptor = results.descriptor
    inp = results.input
    from boutiques.importer import Importer
    importer = Importer(descriptor)
    if results.type == "0.4":
        importer.upgrade_04(inp)
    elif results.type == "bids":
        importer.import_bids(inp)


def exporter(*params):
    parser = ArgumentParser("Export Boutiques descriptor to other formats.")
    parser.add_argument("type", help="Type of export we are performing."
                        " For carmin, pipeline id is set to the output"
                        " file name.",
                        choices=["carmin"])
    parser.add_argument("descriptor", help="Boutiques descriptor to export.")
    parser.add_argument("output", help="Output file where to write the"
                        " converted descriptor.")
    results = parser.parse_args(params)

    descriptor = results.descriptor
    output = results.output

    bosh(["validate", results.descriptor])

    from boutiques.exporter import Exporter
    exporter = Exporter(descriptor)
    if results.type == "carmin":
        exporter.carmin(output)


def publish(*params):
    parser = ArgumentParser("Boutiques publisher",
                            description="A publisher of Boutiques tools"
                            " in Zenodo (http://zenodo.org). Requires "
                            "a Zenodo access token, see "
                            "http://developers.zenodo.org/#authentication.")
    parser.add_argument("boutiques_descriptor", action="store",
                        help="local path of the "
                        " Boutiques descriptor to publish.")
    parser.add_argument("creator", action="store",
                        help="creator to use in Zenodo metadata.")
    parser.add_argument("affiliation", action="store",
                        help="affiliation to use in Zenodo metadata.")
    parser.add_argument("--sandbox", action="store_true",
                        help="publish to Zenodo's sandbox instead of "
                        "production server. Recommended for tests.")
    parser.add_argument("--zenodo-token", action="store",
                        help="Zenodo API token to use for authentication. "
                        "If not used, token will be read from configuration "
                        "file or requested interactively.")
    parser.add_argument("--no-int", '-y', action="store_true",
                        help="disable interactive input.")
    parser.add_argument("-v", "--verbose", action="store_true",
                        help="print information messages.")

    results = parser.parse_args(params)

    from boutiques.publisher import Publisher
    publisher = Publisher(results.boutiques_descriptor,
                          results.creator,
                          results.affiliation,
                          results.verbose,
                          results.sandbox,
                          results.no_int,
                          results.zenodo_token).publish()


def invocation(*params):
    parser = ArgumentParser("Creates invocation schema and validates"
                            " invocations. Uses descriptor's invocation"
                            " schema if it exists, otherwise creates one.")
    parser.add_argument("descriptor", action="store",
                        help="The Boutiques descriptor.")
    parser.add_argument("-i", "--invocation", action="store",
                        help="Input values in a JSON file to be"
                        " validated against "
                        "the invocation schema.")
    parser.add_argument("-w", "--write-schema", action="store_true",
                        help="If descriptor doesn't have an invocation "
                        "schema, creates one and writes it to the descriptor"
                        " file ")
    result = parser.parse_args(params)

    validate(result.descriptor)
    if result.invocation:
        data = json.loads(open(result.invocation).read())

    descriptor = json.loads(open(result.descriptor).read())
    if descriptor.get("invocation-schema"):
        invSchema = descriptor.get("invocation-schema")
    else:
        from boutiques.invocationSchemaHandler import generateInvocationSchema
        invSchema = generateInvocationSchema(descriptor)
        if result.write_schema:
            descriptor["invocation-schema"] = invSchema
            with open(result.descriptor, "w") as f:
                f.write(json.dumps(descriptor, indent=4, sort_keys=True))
    if result.invocation:
        from boutiques.invocationSchemaHandler import validateSchema
        validateSchema(invSchema, data)


def evaluate(*params):
    parser = ArgumentParser("Evaluates parameter values for a descriptor"
                            " and invocation")
    parser.add_argument("descriptor", action="store",
                        help="The Boutiques descriptor.")
    parser.add_argument("invocation", action="store",
                        help="Input JSON complying to invocation.")
    parser.add_argument("query", action="store", nargs="*",
                        help="The query to be performed. Simply request keys "
                        "from the descriptor (i.e. output-files), and chain "
                        "together queries (i.e. id=myfile or optional=false) "
                        "slashes between them and commas connecting them. "
                        "(i.e. output-files/optional=false,id=myfile). "
                        "Perform multiple queries by separating them with a "
                        "space.")
    result = parser.parse_args(params)

    # Generate object that will parse the invocation and descriptor
    from boutiques.localExec import LocalExecutor
    executor = LocalExecutor(result.descriptor,
                             {"forcePathType": True,
                              "destroyTempScripts": True,
                              "changeUser": True})
    executor.readInput(result.invocation)

    from boutiques.evaluate import evaluateEngine
    query_results = []
    for query in result.query:
        query_results += [evaluateEngine(executor, query)]
    return query_results[0] if len(query_results) == 1 else query_results


def test(*params):

    parser = ArgumentParser("Perform all the tests defined within the"
                            " given descriptor")
    parser.add_argument("descriptor", action="store", help="The Boutiques"
                        " descriptor.")
    result = parser.parse_args(params)

    # Generation of the invocation schema (and descriptor validation).
    invocation(result.descriptor)

    # Extraction of all the invocations defined for the test-cases.
    with open(result.descriptor) as fhandle:
        # might just need to be `fhandle` in this context, but not sure
        descriptor = json.loads(fhandle.read())

    if (not descriptor.get("tests")):
        # If no tests have been specified, we consider testing successful.
        return 0

    for test in descriptor["tests"]:
        # Create temporary file for the invocation() function.
        invocation_JSON = test["invocation"]
        temp_invocation_JSON = tempfile.NamedTemporaryFile(suffix=".json",
                                                           delete=False)
        temp_invocation_JSON.write(json.dumps(invocation_JSON).encode())
        temp_invocation_JSON.seek(0)
        # Check if the invocation is valid.
        invocation(result.descriptor, "--invocation", temp_invocation_JSON.name)
        # Destroy the temporary file.
        temp_invocation_JSON.close()

    # Invocations have been properly validated. We can launch the actual tests.

    test_path = op.join(op.dirname(op.realpath(__file__)), "test.py")
    return pytest.main([test_path, "--descriptor", result.descriptor])


def bosh(args=None):
    parser = ArgumentParser(description="Driver for Bosh functions",
                            add_help=False)
    parser.add_argument("function", action="store", nargs="?",
                        help="The tool within boutiques/bosh you wish to run. "
                        "Validate: validates an existing boutiques descriptor."
                        "Exec: launches or simulates an execution given a "
                        "descriptor and a set of inputs. Import: creates a "
                        "descriptor for a BIDS app or updates a descriptor "
                        "from an older version of the schema. Export: exports a"
                        "descriptor to other formats. Publish: creates"
                        "an entry in Zenodo for the descriptor and tool."
                        "Invocation: generates the invocation schema for a "
                        "given descriptor. Eval: given an invocation and a "
                        "descriptor, queries execution properties."
                        "Test: run pytest on a descriptor detailing tests",
                        choices=["validate", "exec", "import", "export",
                                 "publish", "invocation", "evaluate", "test"])

    parser.add_argument("--help", "-h", action="store_true",
                        help="show this help message and exit")

    args, params = parser.parse_known_args(args)
    func = args.function
    params += ["--help"] if args.help is True else []

    # Returns True if bosh was called from the CLI
    def runs_as_cli():
        return os.path.basename(sys.argv[0]) == "bosh"

    try:
        if func == "validate":
            out = validate(*params)
            return out
        elif func == "exec":
            out = execute(*params)
            return out[2]
        elif func == "import":
            out = importer(*params)
            return out
        elif func == "export":
            out = exporter(*params)
            return out
        elif func == "publish":
            out = publish(*params)
            return out
        elif func == "invocation":
            out = invocation(*params)
            return out
        elif func == "evaluate":
            out = evaluate(*params)
            return out
        elif func == "test":
            out = test(*params)
            return out
        else:
            parser.print_help()
            raise SystemExit

    except (ZenodoError,
            DescriptorValidationError,
            InvocationValidationError,
            ToolOutputNotFoundError) as e:
        # We don't want to raise an exception when function is called
        # from CLI.'
        if runs_as_cli():
            print(e)
            return 99  # Note: this conflicts with tool error codes.
        raise e



import os.path as op
from boutiques import __file__ as bfile
import boutiques as bosh
import hashlib
import pytest


def compute_md5(filename):
    with open(filename, 'rb') as fhandle:
        return hashlib.md5(open(filename, 'rb').read()).hexdigest()


def test(descriptor, test, invocation):
    # Run pipeline.
    stdout, stderr, exit_code, err_msg = bosh.execute("launch",
                                                      descriptor,
                                                      invocation.name)

    # Choose appropriate assertion scenario
    assertions = test["assertions"]
    if "exit-code" in assertions:
        assert exit_code == assertions["exit-code"]

    if "output-files" in assertions:

        # Acquiring a hash map of output ids mapping to output file paths.
        outputted = bosh.evaluate(descriptor, invocation.name, "output-files/")

        for output_file in assertions["output-files"]:

            file_path = outputted[output_file["id"]]
            assert op.exists(file_path)

            # Optionaly, an md5 reference may have been specified
            if "md5-reference" in output_file:

                # MD5 checksum comparaison
                output_md5 = compute_md5(file_path)
                reference_md5 = output_file["md5-reference"]
                assert output_md5 == reference_md5



#!/usr/bin/env python

import argparse
import os
import sys
import json
import random as rnd
import string
import math
import random
import subprocess
import time
import pwd
import os.path as op


class ToolOutputNotFoundError(Exception):
    pass


# Executor class
class LocalExecutor(object):
    """
    This class represents a json descriptor of a tool, and can execute
    various tasks related to it. It is constructed first via an
    input json descriptor file, which is held in the desc_dict field.

    An input can be added to it via the in_dict field, a dictionary from
    param ids to values. The in_dict field should only be modified
    via the public readInput method, which can either take
    a file (json or csv) or a string written in the command line.
    The field is always validated by checking the input parameters
     with respect to the descriptor.

    Other public methods include:
        execute - attempts to execute the tool described by the descriptor
            based on the current input (in in_dict)
        printCmdLine - simply prints the generated command line
            based on the current input values
        generateRandomParams - fills in_dict with random values
            (following the constraints of the descriptor schema)
    """

    # Constructor
    def __init__(self, desc, options={}):
        # Initial parameters
        self.desc_path = desc    # Save descriptor path
        self.errs = []        # Empty errors holder
        self.debug = False  # debug mode (also use python -u)
        # Parse JSON descriptor
        with open(desc, 'r') as descriptor:
            self.desc_dict = json.loads(descriptor.read())
        # Helpers Functions
        # The set of input parameters from the json descriptor
        self.inputs = self.desc_dict['inputs']  # Struct: [{id:}..,{id:}]
        # The set of output parameters from the json descriptor
        self.outputs = self.desc_dict['output-files']  # Struct: [{id:}..,{id:}]
        # The set of parameter groups, according to the json descriptor
        self.groups = []
        if 'groups' in list(self.desc_dict.keys()):
            self.groups = self.desc_dict['groups']

        # Container-image Options
        self.con = self.desc_dict.get('container-image')
        self.launchDir = None
        if self.con is not None:
            self.con.get('working-directory')

        # Extra Options
        # Include: forcePathType and destroyTempScripts
        for option in list(options.keys()):
            setattr(self, option, options.get(option))
        # Container Implementation check
        conEngines = ['docker', 'singularity']
        if (self.con is not None) and self.con['type'] not in conEngines:
                msg = "Other container types than {0} (e.g. {1})"\
                      " are not yet supported"
                raise ValueError(msg.format(", ".join(conEngines),
                                            self.con['type']))

    # Retrieves the parameter corresponding to the given id
    def byId(self, n):
        return [v for v in self.inputs+self.outputs if v['id'] == n][0]

    # Retrieves the group corresponding to the given id
    def byGid(self, g):
        return [v for v in self.groups if v['id'] == g][0]

    # Retrieves the value of a field of an input
    # from the descriptor. Returns None if not present.
    def safeGet(self, i, k):
        if k not in list(self.byId(i).keys()):
            return None
        return self.byId(i)[k]

    # Retrieves the value of a field of a group from
    # the descriptor. Returns None if not present.
    def safeGrpGet(self, g, k):
        if k not in list(self.byGid(g).keys()):
            return None
        return self.byGid(g)[k]

    # Retrieves the group a given parameter id belongs to;
    # otherwise, returns None
    def assocGrp(self, i):
        return ([g for g in self.groups if i in g["members"]] or [None])[0]

    # Returns the required inputs of a given input id, or the empty string
    def reqsOf(self, t):
        return self.safeGet(t, "requires-inputs") or []

    # Attempt local execution of the command line
    # generated from the input values
    def execute(self, mount_strings):
        '''
        The execute method runs the generated command line
        (from either generateRandomParams or readInput)
        If docker is specified, it will attempt to use it, instead
        of local execution.
        After execution, it checks for output file existence.
        '''
        command, exit_code, con = self.cmdLine[0], None, self.con or {}
        print('Attempting execution of command:\n\t' + command)
        # Check for Container image
        conType, conImage = con.get('type'), con.get('image'),
        conIndex = con.get("index")
        conIsPresent = (conImage is not None)
        # Export environment variables, if they are specified in the descriptor
        envVars = {}
        if 'environment-variables' in list(self.desc_dict.keys()):
            variables = [(p['name'], p['value']) for p in
                         self.desc_dict['environment-variables']]
            for (envVarName, envVarValue) in variables:
                os.environ[envVarName] = envVarValue
                envVars[envVarName] = envVarValue
        # Container script constant name
        # Note that docker/singularity cannot do a local volume
        # mount of files starting with a '.', hence this one does not
        millitime = int(time.time()*1000)
        dsname = ('temp-' +
                  str(random.SystemRandom().randint(0, int(millitime))) +
                  "-" + str(millitime) + '.localExec.boshjob.sh')
        dsname = op.realpath(dsname)
        # If container is present, alter the command template accordingly
        if conIsPresent:
            if conType == 'docker':
                # Pull the docker image
                if self._localExecute("docker pull " + str(conImage))[1]:
                    print("Container not found online,  trying local copy")
            elif conType == 'singularity':
                if not conIndex:
                    conIndex = "shub://"
                elif not conIndex.endswith("://"):
                    conIndex = conIndex + "://"
                conName = conImage.replace("/", "-").replace(":", "-") + ".simg"

                if conName not in os.listdir('./'):
                    pull_location = "\"{0}\" {1}{2}".format(conName,
                                                            conIndex,
                                                            conImage)
                    print("Container image ({0}) not found in current"
                          " working directory,"
                          " pulling from {1}".format(conName, pull_location))
                    # Pull the singularity image
                    if self._localExecute("singularity pull --name " +
                                          pull_location)[1]:
                        print("Could not pull image.")
                        sys.exit(1)
                else:
                    print("Using local container image: {0}".format(conName))
                conName = op.abspath(conName)
            else:
                print('Unrecognized container type: \"%s\"' % conType)
                sys.exit(1)
            # Generate command script
            uname, uid = pwd.getpwuid(os.getuid())[0], str(os.getuid())
            # Adds the user to the container before executing
            # the templated command line
            userchange = '' if not self.changeUser else ("useradd --uid " +
                                                         uid + ' ' + uname +
                                                         "\n")
            # If --changeUser was desired, run with su so that
            # any output files are owned by the user instead of root
            # Get the supported shell by the docker or singularity
            if self.changeUser:
                command = 'su ' + uname + ' -c ' + "\"{0}\"".format(command)
            cmdString = "#!/bin/sh -l\n" + userchange + str(command)
            with open(dsname, "w") as scrFile:
                scrFile.write(cmdString)
            # Ensure the script is executable
            self._localExecute("chmod 755 " + dsname)
            # Prepare extra environment variables
            envString = ""
            if envVars:
                for (key, val) in list(envVars.items()):
                    envString += "SINGULARITYENV_{0}='{1}' ".format(key, val)
            # Change launch (working) directory if desired
            launchDir = self.launchDir
            if launchDir is None:
                launchDir = op.realpath('./')
            launchDir = op.realpath(launchDir)
            # Run it in docker
            mount_strings = [] if not mount_strings else mount_strings
            mount_strings = [op.realpath(m.split(":")[0])+":"+m.split(":")[1]
                             for m in mount_strings]
            mount_strings.append(op.realpath('./') + ':' + launchDir)
            if conType == 'docker':
                envString = " "
                if envVars:
                    for (key, val) in list(envVars.items()):
                        envString += " -e {0}='{1}' ".format(key, val)
                # export mounts to docker string
                docker_mounts = " -v ".join(m for m in mount_strings)
                dcmd = ('docker run --entrypoint=/bin/sh --rm' + envString +
                        ' -v ' + docker_mounts + ' -w ' + launchDir + ' ' +
                        str(conImage) + ' ' + dsname)
            elif conType == 'singularity':
                envString = ""
                if envVars:
                    for (key, val) in list(envVars.items()):
                        envString += "SINGULARITYENV_{0}='{1}' ".format(key,
                                                                        val)
                # TODO: Singularity 2.4.6 default configuration binds: /proc,
                # /sys, /dev, ${HOME}, /tmp, /var/tmp, /etc/localtime, and
                # /etc/hosts. This means that any path down-stream shouldn't
                # be bound on the command-line, as this will currently raise
                # an exception. See:
                #   https://github.com/singularityware/singularity/issues/1469
                #
                # Previous bind string:
                #   singularity_mounts = " -B ".join(m for m in mount_strings)

                def_mounts = ["/proc", "/sys", "/dev", "/tmp", "/var/tmp",
                              "/etc/localtime", "/etc/hosts",
                              op.realpath(op.expanduser('~')),
                              op.expanduser('~')]

                # Ensures the set of paths provided has no overlap
                compaths = list()
                for idxm, m in enumerate(mount_strings):
                    for n in mount_strings[idxm:]:
                        if n != m:
                            tmp = op.dirname(op.commonprefix([n, m]))
                            if tmp != '/':
                                compaths += [tmp]
                    if not any(m.startswith(c) for c in compaths):
                        compaths += [m]
                mount_strings = set(compaths)

                # Only adds mount points for those not already included
                singularity_mounts = ""
                for m in mount_strings:
                    if not any(d in m for d in def_mounts):
                        singularity_mounts += "-B {0} ".format(m)

                dcmd = (envString + 'singularity exec --cleanenv ' +
                        singularity_mounts + ' -W ' + launchDir + ' ' +
                        str(conName) + ' ' + dsname)
            else:
                print('Unrecognized container type: \"%s\"' % conType)
                sys.exit(1)
            print('Executing via: ' + dcmd)
            (stdout, stderr), exit_code = self._localExecute(dcmd)
        # Otherwise, just run command locally
        else:
            (stdout, stderr), exit_code = self._localExecute(command)
        # Report exit status
        print('---/* Begin program output */---')
        if stdout != '':
            print(stdout.decode('utf-8'))
        print('---/* End program output */---\nCompleted execution'
              ' (exit code: ' + str(exit_code) + ')')
        time.sleep(0.5)  # Give the OS a (half) second to finish writing
        # Destroy temporary docker script, if desired.
        # By default, keep the script so the dev can look at it.
        if conIsPresent and self.destroyTempScripts:
            if os.path.isfile(dsname):
                os.remove(dsname)
        # Check for output files (note: the path-template
        # can contain value-keys)
        print('Looking for output files:')
        for outfile in self.desc_dict['output-files']:
            outFileName = self.out_dict[outfile['id']]
            # Look for the target file
            exists = os.path.exists(outFileName)
            # Note whether it could be found or not
            if 'optional' in list(outfile.keys()):
                isOptional = outfile['optional']
            else:
                isOptional = False
            s1 = 'Optional' if isOptional else 'Required'
            s2 = '' if exists else 'not '
            message = (s1 + " output file \'" + outfile['name'] +
                       "\' was " + s2 + "found at " + outFileName)
            if (not isOptional and not exists):
                raise ToolOutputNotFoundError(message)
            else:
                print("\t {0}".format(message))
        desc_err = ''
        if 'error-codes' in list(self.desc_dict.keys()):
            for err_elem in self.desc_dict['error-codes']:
                if err_elem['code'] == exit_code:
                        desc_err = err_elem['description']
                        break
        error_msg = ''
        if stderr != b'':
            error_msg = 'Execution ERR ({0}): {1}'.format(exit_code, stderr)
        if desc_err != '':
            error_msg += '{0}{1} ERR ({2}): {3}'.format('\n' if
                                                        error_msg != '' else
                                                        '',
                                                        self.desc_dict['name'],
                                                        exit_code, desc_err)

        return stdout, stderr, exit_code, error_msg

    # Private method that attempts to locally execute the given
    # command. Returns the exit code.
    def _localExecute(self, command):
        # Note: invokes the command through the shell
        # (potential injection dangers)
        try:
            process = subprocess.Popen(command, shell=True,
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.PIPE)

        except OSError as e:
            sys.stderr.write('OS Error during attempted execution!')
            raise e
        except ValueError as e:
            sys.stderr.write('Input Value Error during attempted execution!')
            raise e
        else:
            return process.communicate(), process.returncode

    # Private method to generate a random input parameter set that follows
    # the constraints from the json descriptor
    # This method fills in the in_dict field of the object
    # with constrained random values
    def _randomFillInDict(self):
        # Private helper functions for filling the dictionary
        # Helpers for generating random numbers, strings, etc...
        # Note: uses-absolute-path is satisfied for files by the
        # automatic replacement in the _validateDict
        # nd = number of random characters to use in generating strings,
        # nl = max number of random list items
        nd, nl = 2, 5

        def randDigs():
            # Generate random string of digits
            return ''.join(rnd.choice(string.digits)
                           for _ in range(nd))

        def randFile():
            return ('f_' + randDigs() +
                    rnd.choice(['.csv', '.tex', '.j',
                                '.cpp', '.m', '.mnc',
                                '.nii.gz']))

        def randStr():
            return 'str_' + ''.join(rnd.choice(string.digits +
                                    string.ascii_letters)
                                    for _ in range(nd))

        # A function for generating a number type parameter input
        # p is a dictionary object corresponding to a parameter
        # description in the json descriptor
        # E.g. if p had no constraints from the json descriptor,
        # the output would be a float in [defaultMin,defaultMax]
        #            if p had "integer": true, "minimum": 7,
        # "maximum": 9, the output would be an int in [7,9]
        def randNum(p):
            param_id, defaultMin, defaultMax = p['id'], -50, 50
            # Check if the input parameter should be an int
            isInt = self.safeGet(param_id, 'integer')

            def roundTowardsZero(x):
                return int(math.copysign(1, x) * int(abs(x)))
            # Assign random values to min and max,
            # unless they have been specified
            minv = self.safeGet(param_id, 'minimum')
            maxv = self.safeGet(param_id, 'maximum')
            if minv is None and maxv is None:
                minv, maxv = defaultMin, defaultMax
            elif minv is None and not (maxv is None):
                minv = maxv + defaultMin
            elif not (minv is None) and maxv is None:
                maxv = minv + defaultMax
            # Coerce the min/max to the proper number type
            if isInt:
                minv, maxv = roundTowardsZero(minv), roundTowardsZero(maxv)
            else:
                minv, maxv = float(minv), float(maxv)
            # Apply exclusive boundary constraints, if any
            if self.safeGet(param_id, 'exclusive-minimum'):
                minv += (1 if isInt else 0.0001)
            if self.safeGet(param_id, 'exclusive-maximum'):
                maxv -= (1 if isInt else 0.0001)
            # Returns a random int or a random float, depending on the type of p
            return (rnd.randint(minv, maxv)
                    if isInt else round(rnd.uniform(minv, maxv), nd))

        # Generate a random parameter value based on the input
        # type (where prm \in self.inputs)
        def paramSingle(prm):
            if self.safeGet(prm['id'], 'value-choices'):
                return rnd.choice(self.safeGet(prm['id'], 'value-choices'))
            if prm['type'] == 'String':
                return randStr()
            if prm['type'] == 'Number':
                return randNum(prm)
            if prm['type'] == 'Flag':
                return rnd.choice(['true', 'false'])
            if prm['type'] == 'File':
                return randFile()

        # For this function, given prm (a parameter description),
        # a parameter value is generated
        # If prm is a list, a sequence of outputs is generated;
        # otherwise, a single value is returned
        def makeParam(prm):
            mn = self.safeGet(prm['id'], 'min-list-entries') or 2
            mx = self.safeGet(prm['id'], 'max-list-entries') or nl
            isList = self.safeGet(prm['id'], 'list') or False
            return ' '.join(str(paramSingle(prm)) for _ in
                            range(rnd.randint(mn, mx))
                            ) if isList else paramSingle(prm)

        # Returns a list of the ids of parameters that
        # disable the input parameter
        def disablersOf(inParam):
            return [disabler[0] for disabler in
                    [disabler for disabler in
                     [(prm['id'], self.safeGet(prm['id'],
                                               'disables-inputs') or [])
                      for prm in self.inputs] if inParam['id'] in disabler[1]]]

        # Returns the list of mutually requiring parameters of the target
        def mutReqs(targetParam):
            return [self.byId(mutualReq[0]) for mutualReq in
                    [possibleMutReq for possibleMutReq in
                     [(reqOfTarg, self.reqsOf(reqOfTarg)) for reqOfTarg in
                         self.reqsOf(targetParam['id'])]
                     if targetParam['id'] in possibleMutReq[1]]]

        # Returns whether targ (an input parameter) has a
        # value or is allowed to have one
        def isOrCanBeFilled(targ):
            # If it is already filled in, report so
            if targ['id'] in list(self.in_dict.keys()):
                return True
            # If a disabler or a disabled target is already active,
            # it cannot be filled
            for d in disablersOf(targ) + (self.safeGet(
                                                 targ['id'],
                                                 'disables-inputs') or []):
                if d in list(self.in_dict.keys()):
                    return False
            # If at least one non-mutual requirement has
            # not been met, it cannot be filled
            for r in self.reqsOf(targ['id']):
                if r not in self.in_dict:  # If a requirement is not present
                    # and it is not mutually required
                    if targ['id'] not in self.reqsOf(r):
                        return False
            # If it is in a mutex group with one target already chosen,
            # it cannot be filled
            # Get the group that the target belongs to, if any
            g = self.assocGrp(targ['id'])
            if (g is not None) and self.safeGrpGet(g['id'],
                                                   'mutually-exclusive'):
                if len([x for x in g['members']
                        if x in list(self.in_dict.keys())]) > 0:
                    return False
            return True

        # Handle the mutual requirement case by breadth first search in
        # the graph of mutual requirements. Essentially a graph is built,
        # starting from targ (the target input parameter), where nodes
        # are input parameters and edges are a mutual requirement
        # relation between nodes. BFS is used to check every node, to see
        # if can be (or has been) given a value. If all the nodes are
        # permitted to be given values, then they are all added at once;
        # if even one of them cannot be given a value (e.g. it has an
        # active disabler) then none of the graph members can be added
        # and so we just return false. # Input: an input parameter from
        # which to start building the graph Output: Returns False if at
        # least one of the mutual requirements cannot be met Returns a
        # list of params to fill if all of them can be met (or [targ.id]
        # if it has no mutReqs)
        def checkMutualRequirements(targ):
            checked, toCheck = [], [targ]
            while len(toCheck) > 0:
                current = toCheck.pop()
                checked.append(current)
                if not isOrCanBeFilled(current):
                    return False
                for mutreq in mutReqs(current):
                    if not mutreq['id'] in [c['id'] for c in checked]:
                        toCheck.append(mutreq)
            return checked

        # Start actual dictionary filling part
        # Clear the dictionary
        self.in_dict = {}
        # Fill in the required parameters
        for reqp in [r for r in self.inputs if not r.get('optional')]:
            print(reqp['id'])
            self.in_dict[reqp['id']] = makeParam(reqp)
        # Fill in a random choice for each one-is-required group
        for grp in [g for g in self.groups
                    if self.safeGrpGet(g['id'], 'one-is-required')]:
            # Loop to choose an allowed value,
            # in case a previous choice disabled that one
            while True:
                # Pick a random parameter
                choice = self.byId(rnd.choice(grp['members']))
                # see if it and its mutual requirements can be filled
                res = checkMutualRequirements(choice)
                if res is False:
                    # Try again if the chosen group member is not permissable
                    continue
                for r in res:
                    self.in_dict[r['id']] = makeParam(r)
                break  # If we were allowed to add a parameter, we can stop
        # Choose a random number of times to try to fill optional inputs
        opts = [p for p in self.inputs
                if self.safeGet(p['id'], '') in [None, True]]
        # Loop a random number of times, each time
        #  attempting to fill a random parameter
        for _ in range(rnd.randint(int(len(opts) / 2 + 1), len(opts) * 2)):
            targ = rnd.choice(opts)  # Choose an optional output
            # If it is already filled in, continue
            if targ['id'] in list(self.in_dict.keys()):
                continue
            # If it is a prohibited option, continue
            # (isFilled case handled above)
            if not isOrCanBeFilled(targ):
                continue
            # Now we handle the mutual requirements case. This is a little
            # more complex because a mutual requirement
            # of targ can have its own mutual requirements, ad nauseam.
            # We need to look at all of them recursively and either
            # fill all of them in (i.e. at once) or none of them
            # (e.g. if one of them is disabled by some other param).
            result = checkMutualRequirements(targ)
            # Leave if the mutreqs cannot be satisfied
            if result is False:
                continue
            # Fill in the target(s) otherwise
            for r in result:
                self.in_dict[r['id']] = makeParam(r)

    # Function to generate random parameter values
    # This fills the in_dict with random values, validates the input,
    # and generates the appropriate command line
    def generateRandomParams(self, n):

        '''
        The generateRandomParams method fills the in_dict field
        with randomly generated values following the schema.
        It then generates command line strings based on these
        values (more than 1 if -n was given).
        '''

        self.cmdLine = []
        for i in range(0, n):
            # Set in_dict with random values
            self._randomFillInDict()
            # Look at generated input, if debugging
            if self.debug:
                print("Input: " + str(self.in_dict))
            # Check results (as much as possible)
            try:
                self._validateDict()
            # If an error occurs, print out the problems already
            # encountered before blowing up
            except Exception:  # Avoid catching BaseExceptions like SystemExit
                sys.stderr.write("An error occurred in validation\n"
                                 "Previously saved issues\n")
                for err in self.errs:
                    sys.stderr.write("\t" + str(err) + "\n")
                raise  # Pass on (throw) the caught exception
            # Add new command line
            self.cmdLine.append(self._generateCmdLineFromInDict())

    # Read in parameter input file or string
    def readInput(self, infile):

        '''
        The readInput method sets the in_dict field of the executor
        object, based on a fixed input.
        It then generates a command line based on the input.

        infile: either the inputs in a file or
        the command-line string (from -s).
        stringInput: a boolean as to whether the method has
        been given a string or a file.
        '''

        # Quick check that the descriptor has already been read in
        assert self.desc_dict is not None
        with open(infile, 'r') as inparams:
            self.in_dict = json.loads(inparams.read())
        # Input dictionary
        if self.debug:
            print("Input: " + str(self.in_dict))
        # Fix special flag case: flags given the false value
        # are treated as non-existent
        toRm = []
        for inprm in self.in_dict:
            if (str(self.in_dict[inprm]).lower() == 'false'
               and self.byId(inprm)['type'] == 'Flag'):
                    toRm.append(inprm)
            elif (self.byId(inprm)['type'] == 'Flag'
                  and self.in_dict[inprm] is True):
                # Fix json inputs using bools instead of strings
                self.in_dict[inprm] = "true"
        for r in toRm:
            del self.in_dict[r]
        # Add default values for required parameters,
        # if no value has been given
        for input in [s for s in self.inputs
                      if s.get("default-value") is not None
                      and not s.get("optional")]:
            if self.in_dict.get(input['id']) is None:
                df = input.get("default-value")
                if not input['type'] == 'Flag':
                    self.in_dict[input['id']] = df
                else:
                    self.in_dict[input['id']] = str(df)
        # Check results (as much as possible)
        try:
            pass  # self._validateDict()
        except Exception:  # Avoid catching BaseExceptions like SystemExit
            sys.stderr.write("An error occurred in validation\n"
                             "Previously saved issues\n")
            for err in self.errs:
                # Write any errors we found
                sys.stderr.write("\t" + str(err) + "\n")
            raise  # Raise the exception that caused failure
        # Build and save output command line (as a single-entry list)
        self.cmdLine = [self._generateCmdLineFromInDict()]

    # Private method to replace the keys in template by input and output
    # values. Input and output values are looked up in self.in_dict and
    # self.out_dict
    # * if useFlags is true, keys will be replaced by flag+flag-separator+value
    # * if unfoundKeys is "remove", unfound keys will be replaced by ""
    # * if unfoundKeys is "clear" then the template is cleared if it has
    #     unfound keys (useful for configuration files)
    # * before being substituted, the keys will be stripped from all
    #    the strings in strippedExtensions
    def _replaceKeysInTemplate(self, template,
                               useFlags=False, unfoundKeys="remove",
                               strippedExtensions=[]):
            # Concatenate input and output dictionaries
            in_out_dict = dict(self.in_dict)
            in_out_dict.update(self.out_dict)
            # Go through all the keys
            for paramId in [x['id'] for x in self.inputs + self.outputs]:
                clk = self.safeGet(paramId, 'value-key')
                if clk is None:
                    continue
                if paramId in list(in_out_dict.keys()):  # param has a value
                    val = in_out_dict[paramId]
                    if type(val) is list:
                        s_val = ""
                        for x in val:
                            s_val += str(x) + " "
                        val = s_val
                    else:
                            val = str(val)
                    # Add flags and separator if necessary
                    if useFlags:
                        flag = self.safeGet(paramId, 'command-line-flag') or ''
                        sep = self.safeGet(paramId,
                                           'command-line-flag-separator') or ' '
                        val = flag + sep + val
                        # special case for flag-type inputs
                        if self.safeGet(paramId, 'type') == 'Flag':
                            val = '' if val.lower() == 'false' else flag
                    # Remove file extensions from input value
                    for extension in strippedExtensions:
                        val = val.replace(extension, "")
                    template = template.replace(clk, val)
                else:  # param has no value
                    if unfoundKeys == "remove":
                        template = template.replace(clk, '')
                    elif unfoundKeys == "clear":
                        if clk in template:
                            return ""
            return template

    # Private method to generate output file names.
    # Output file names will be put in self.out_dict.
    def _generateOutputFileNames(self):
        if not hasattr(self, 'out_dict'):
            # a dictionary that will contain the output file names
            self.out_dict = {}
        for outputId in [x['id'] for x in self.outputs]:
            # Initialize file name with path template or existing value
            if outputId in list(self.out_dict.keys()):
                outputFileName = self.out_dict[outputId]
            else:
                outputFileName = self.safeGet(outputId, 'path-template')
            strippedExtensions = self.safeGet(
                                        outputId,
                                        "path-template-stripped-extensions")
            if strippedExtensions is None:
                strippedExtensions = []
            # We keep the unfound keys because they will be
            # substituted in a second call to the method in case
            # they are output keys
            outputFileName = self._replaceKeysInTemplate(outputFileName,
                                                         False,
                                                         "keep",
                                                         strippedExtensions)
            if self.safeGet(outputId, 'uses-absolute-path'):
                outputFileName = os.path.abspath(outputFileName)
            self.out_dict[outputId] = outputFileName

    # Private method to write configuration files
    # Configuration files are output files that have a file-template
    def _writeConfigurationFiles(self):
        for outputId in [x['id'] for x in self.outputs]:
            fileTemplate = self.safeGet(outputId, 'file-template')
            if fileTemplate is None:
                continue  # this is not a configuration file
            strippedExtensions = self.safeGet(
                                        outputId,
                                        "path-template-stripped-extensions")
            if strippedExtensions is None:
                strippedExtensions = []
            # We substitute the keys line by line so that we can
            # clear the lines that have keys with no value
            # (undefined optional params)
            newTemplate = []
            for line in fileTemplate:
                newTemplate.append(self._replaceKeysInTemplate(
                                                line,
                                                False, "clear",
                                                strippedExtensions))
            template = "\n".join(newTemplate)
            # Write the configuration file
            fileName = self.out_dict[outputId]
            file = open(fileName, 'w')
            file.write(template)
            file.close()

    # Private method to build the actual command line by substitution,
    # using the input data
    def _generateCmdLineFromInDict(self):
        # Genrate output file names
        self._generateOutputFileNames()
        # it is required to call the method twice in case path
        # templates contain output keys
        self._generateOutputFileNames()
        # Write configuration files
        self._writeConfigurationFiles()
        # Get the command line template
        template = self.desc_dict['command-line']
        # Substitute every given value into the template
        # (incl. flags, flag-seps, ...)
        template = self._replaceKeysInTemplate(template, True, "remove")
        # Return substituted command line
        return template

    # Print the command line result
    def printCmdLine(self):
        print("Generated Command" +
              ('s' if len(self.cmdLine) > 1 else '') + ':')
        for cmd in self.cmdLine:
            print(cmd)

    # Private method for validating input parameters
    def _validateDict(self):
        # Holder for errors
        self.errs = []

        # Return whether s is a proper number; if intCheck is true,
        # also check if it is an int
        def isNumber(s, intCheck=False):
            try:
                int(s) if intCheck else float(s)
                return True
            except ValueError:
                return False
        # Check individual inputs
        for key in self.in_dict:
            isList = self.safeGet(key, "list")
            # Get current value and schema descriptor properties
            val, targ = self.in_dict[key], self.byId(key)

            # A little closure helper to check if input values satisfy
            # the json descriptor's constraints
            # Checks whether 'value' is appropriate for parameter
            # 'input' by running 'isGood' on it
            # If the input parameter is bad, it adds 'msg' to the list of errors
            def check(keyname, isGood, msg, value):  # Checks input values
                # No need to check constraints if they were not specified
                dontCheck = ((keyname not in list(targ.keys()))
                             or (targ[keyname] is False))
                # Keyname = None is a flag to check the type
                if (keyname is not None) and dontCheck:
                    return
                # The input function is used to check whether
                # the input parameter is acceptable
                if not isGood(value, keyname):
                    self.errs.append(key + ' (' + str(value) + ') ' + msg)
            # The id exists as an input and is not duplicate
            if len([k for k in self.inputs if k['id'] == key]) != 1:
                self.errs.append(key+' is an invalid id')
            # Types are correct
            if targ["type"] == "Number":
                # Number type and constraints are not violated
                # (note the lambda safety checks)
                for v in (str(val).split() if isList else [val]):
                    check('minimum', lambda x, y: float(x) >= targ[y],
                          "violates minimum value", v)
                    check('exclusive-minimum',
                          lambda x, y: float(x) > targ['minimum'],
                          "violates exclusive min value", v)
                    check('maximum', lambda x, y: float(x) <= targ[y],
                          "violates maximum value", v)
                    check('exclusive-maximum',
                          lambda x, y: float(x) < targ['maximum'],
                          "violates exclusive max value", v)
                    check('integer', lambda x, y: isNumber(x, True),
                          "violates integer requirement", v)
                    check(None, lambda x, y: isNumber(x), "is not a number", v)
            elif self.safeGet(targ['id'], 'value-choices'):
                # Value is in the list of allowed values
                check('value-choices', lambda x, y: x in targ[y],
                      "is not a valid enum choice", val)
            elif targ["type"] == "Flag":
                # Should be 'true' or 'false' when lower-cased
                # (based on our transformations of the input)
                check(None,
                      lambda x, y: x.lower() in ["true", "false"],
                      "is not a valid flag value", val)
            elif targ["type"] == "File":
                # Check path-type (absolute vs relative)
                if not self.forcePathType:
                    for ftarg in (str(val).split() if isList else [val]):
                        check('uses-absolute-path',
                              lambda x, y: os.path.isabs(x),
                              "is not an absolute path", ftarg)
                else:
                    # Replace incorrectly specified paths if desired
                    replacementFiles = []
                    launchDir = os.getcwd()
                    if self.launchDir is not None:
                        launchDir = self.launchDir
                    for ftarg in (str(val).split() if isList else [val]):
                        # Special case 1: launchdir is specified and we
                        # want to use absolute path
                        # Note: in this case, the pwd is mounted as the
                        # launchdir; we do not attempt to move files if they
                        # will not be mounted, currently
                        # That is, specified files that are not in the pwd or a
                        # subfolder will not be mounted to the container
                        if (targ.get('uses-absolute-path') is True and
                           self.launchDir is not None):
                            # relative path to target, from the pwd
                            relpath = os.path.relpath(ftarg, os.getcwd())
                            # absolute path in the container
                            mountedAbsPath = os.path.join(launchDir, relpath)
                            replacementFiles.append(
                                        os.path.abspath(mountedAbsPath))
                        # If the input uses-absolute-path, replace
                        # the path with its absolute version
                        elif targ.get('uses-absolute-path') is True:
                            replacementFiles.append(os.path.abspath(ftarg))
                    # Replace old val with the new one
                    self.in_dict[key] = " ".join(replacementFiles)
            # List length constraints are satisfied
            if isList:
                check('min-list-entries',
                      lambda x, y: len(x.split()) >= targ[y],
                      "violates min size", val)
            if isList:
                check('max-list-entries',
                      lambda x, y: len(x.split()) <= targ[y],
                      "violates max size", val)
        # Required inputs are present
        for reqId in [v['id'] for v in self.inputs if not v.get('optional')]:
            if reqId not in list(self.in_dict.keys()):
                self.errs.append('Required input ' + str(reqId) +
                                 ' is not present')
        # Disables/requires is satisfied
        for givenVal in [v for v in self.inputs
                         if v['id'] in list(self.in_dict.keys())]:
            # Check that requirements are present
            for r in self.reqsOf(givenVal['id']):
                if r not in list(self.in_dict.keys()):
                    self.errs.append('Input ' + str(givenVal['id']) +
                                     ' is missing requirement '+str(r))
            for d in (givenVal['disables-inputs']
                      if 'disables-inputs' in list(givenVal.keys()) else []):
                # Check if a disabler is present
                if d in list(self.in_dict.keys()):
                    self.errs.append('Input ' + str(d) +
                                     ' should be disabled by ' +
                                     str(givenVal['id']))
        # Group one-is-required/mutex is ok
        for group, mbs in [(x, x["members"]) for x in self.groups]:
            # Check that the set of parameters in mutually
            # exclusive groups have at most one member present
            if (("mutually-exclusive" in list(group.keys())) and
               group["mutually-exclusive"]):
                if len(set.intersection(set(mbs),
                                        set(self.in_dict.keys()))) > 1:
                    self.errs.append('Group ' + str(group["id"]) +
                                     ' is supposed to be mutex')
            # Check that the set of parameters in one-is-required
            # groups have at least one member present
            if (("one-is-required" in list(group.keys())) and
               group["one-is-required"]):
                    if len(set.intersection(set(mbs),
                                            set(self.in_dict.keys()))) < 1:
                        self.errs.append('Group ' + str(group["id"]) +
                                         ' requires one member to be present')
        # Fast-fail if there was a problem with the input parameters
        if len(self.errs) != 0:
            sys.stderr.write("Problems found with prospective input:\n")
            for err in self.errs:
                sys.stderr.write("\t" + err + "\n")
            sys.exit(1)



#!/usr/bin/env python

# Copyright 2015 - 2017:
#   The Royal Institution for the Advancement of Learning McGill University,
#   Centre National de la Recherche Scientifique,
#   University of Southern California,
#   Concordia University
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import simplejson
import os.path as op
from jsonschema import validate, ValidationError
from argparse import ArgumentParser
from boutiques import __file__ as bfile


# An exception class specific to descriptors
class DescriptorValidationError(ValidationError):
    pass


# Main validation module
def validate_descriptor(json_file, **kwargs):
    """
    Validates the Boutiques descriptor against the schema.
    """
    path, fil = op.split(bfile)
    schema_file = op.join(path, "schema", "descriptor.schema.json")

    # Load schema
    with open(schema_file) as fhandle:
        schema = simplejson.load(fhandle)

    # Load descriptor
    with open(json_file) as fhandle:
        try:
            descriptor = simplejson.load(fhandle)
        except simplejson.errors.JSONDecodeError as e:
            raise DescriptorValidationError(str(e))

    # Validate basic JSON schema compliance for descriptor
    # Note: if it fails basic schema compliance we don"t do more checks
    try:
        validate(descriptor, schema)
    except ValidationError as e:
        raise DescriptorValidationError(str(e))

    # Helper get functions
    def safeGet(desc, sec, targ):
        return [item[targ] for item in desc[sec]
                if list(item.keys()).count(targ)]

    def inputGet(s):
        return safeGet(descriptor, "inputs", s)

    def outputGet(s):
        return safeGet(descriptor, "output-files", s)

    def groupGet(s):
        return safeGet(descriptor, "groups", s)

    def inById(i):
        if i in inputGet("id"):
            return descriptor["inputs"][inputGet("id").index(i)]
        return {}

    # Begin looking at Boutiques-specific failures
    errors = []

    clkeys = inputGet("value-key") + outputGet("value-key")
    flattenedTemplates = [y for x in outputGet("file-template") for y in x]
    configFileTemplates = flattenedTemplates + outputGet("path-template")

    cmdline = descriptor["command-line"]

    # Verify that all command-line key appear in the command-line
    msg_template = "   KeyError: \"{0}\" not in command-line or file template"
    errors += [msg_template.format(k)
               for k in clkeys
               if (cmdline.count(k) +
                   " ".join(configFileTemplates).count(k)) < 1]

    # Verify that no key contains another key
    msg_template = "   KeyError: \"{0}\" contains \"{1}\""
    errors += [msg_template.format(key, clkeys[jdx])
               for idx, key in enumerate(clkeys)
               for jdx in range(0, len(clkeys))
               if clkeys[jdx] in key and key != clkeys[jdx]]

    # Verify that all Ids are unique
    inIds, outIds = inputGet("id"), outputGet("id")
    grpIds = groupGet("id") if "groups" in descriptor.keys() else []
    allIds = inIds + outIds + grpIds
    msg_template = "    IdError: \"{0}\" is non-unique"
    for idx, s1 in enumerate(allIds):
        for jdx, s2 in enumerate(allIds):
            if s1 == s2 and idx < jdx:
                errors += [msg_template.format(s1)]
            else:
                errors += []

    # Verify that identical keys only exist if they are both in mutex groups
    msg_template = " MutExError: \"{0}\" belongs to 2+ non exclusive IDs"
    for idx, key in enumerate(clkeys):
        for jdx in range(idx+1, len(clkeys)):
            if clkeys[jdx] == key:
                mids = [inById(mid)["id"] for mid in inIds
                        if inById(mid)["value-key"] == key]
                for idx, grp in enumerate(descriptor.get("groups")):
                    mutex = grp.get("mutually-exclusive")
                    if set(grp["members"]) == set(mids) and not mutex:
                        errors += [msg_template.format(key)]

    # Verify that output files have unique path-templates
    msg_template = ("OutputError: \"{0}\" and \"{1}\" have the same "
                    "path-template")
    for idx, out1 in enumerate(descriptor["output-files"]):
        for jdx, out2 in enumerate(descriptor["output-files"]):
            if out1["path-template"] == out2["path-template"] and jdx > idx:
                errors += [msg_template.format(out1["id"], out2["id"])]
            else:
                errors += []

    # Verify inputs
    for inp in descriptor["inputs"]:

        # Add optional property in case it's not
        # there (default to false as in JSON)
        if "optional" not in inp.keys():
            inp["optional"] = False

        # Verify flag-type inputs (have flags, not required, cannot be lists)
        if inp["type"] == "Flag":
            msg_template = " InputError: \"{0}\" must have a command-line flag"
            if "command-line-flag" not in inp.keys():
                errors += [msg_template.format(inp["id"])]
            else:
                errors += []

            msg_template = " InputError: \"{0}\" is of type Flag,"\
                           " it has to be optional"
            if inp["optional"] is False:
                errors += [msg_template.format(inp["id"])]
            else:
                errors += []

        # Verify number-type inputs min/max are sensible
        elif inp["type"] == "Number":
            msg_template = (" InputError: \"{0}\" cannot have greater"
                            " min ({1}) than max ({2})")
            minn = inp["minimum"] if "minimum" in inp.keys() else -float("Inf")
            maxx = inp["maximum"] if "maximum" in inp.keys() else float("Inf")
            if minn > maxx:
                errors += [msg_template.format(inp["id"], minn, maxx)]
            else:
                errors += []

        # Verify enum-type inputs (at least 1 option, default in set)
        elif "value-choices" in inp.keys():
            msg_template = (" InputError: \"{0}\" must have at least"
                            " one value choice")
            if len(inp["value-choices"]) < 1:
                errors += [msg_template.format(inp["id"])]
            else:
                errors += []

            msg_template = " InputError: \"{0}\" cannot have default"\
                           " value outside its choices"
            if ("default-value" in inp.keys()
               and inp["default-value"] not in inp["value-choices"]):
                    errors += [msg_template.format(inp["id"])]
            else:
                errors += []

        # Verify list-type inputs (min entries less than max,
        # no negative entries (both on min and max)
        if "list" in inp.keys():
            msg_template = (" InputError: \"{0}\" cannot have greater min"
                            " entries ({1}) than max entries ({2})")
            minn = inp.get("min-list-entries") or 0
            maxx = inp.get("max-list-entries") or float("Inf")
            if minn > maxx:
                errors += [msg_template.format(inp["id"], minn, maxx)]
            else:
                errors += []

            msg_template = (" InputError: \"{0}\" cannot have negative min"
                            " entries ({1})")
            errors += [msg_template.format(inp["id"], minn)] if minn < 0 else []

            msg_template = (" InputError: \"{0}\" cannot have non-positive"
                            " max entries ({1})")
            if maxx <= 0:
                errors += [msg_template.format(inp["id"], maxx)]
            else:
                errors += []

        # Verify requires- and disables-inputs (present ids, non-overlapping)
        msg_template = " InputError: \"{0}\" {1}d id \"{2}\" not found"
        for param in ["require", "disable"]:
            if param+"s-inputs" in inp.keys():
                errors += [msg_template.format(inp["id"], param, ids)
                           for ids in inp[param+"s-inputs"]
                           if ids not in inIds]

        if "requires-inputs" in inp.keys() and "disables-inputs" in inp.keys():
            msg_template = " InputError: \"{0}\" requires and disables \"{1}\""
            errors += [msg_template.format(inp["id"], ids1)
                       for ids1 in inp["requires-inputs"]
                       for ids2 in inp["disables-inputs"]
                       if ids1 == ids2]

        # Verify required inputs cannot require or disable other parameters
        if "requires-inputs" in inp.keys() or "disables-inputs" in inp.keys():
            msg_template = (" InputError: \"{0}\" cannot require or"
                            " disable other inputs")
            if not inp["optional"]:
                errors += [msg_template.format(inp["id"])]

    # Verify groups
    for idx, grpid in enumerate(grpIds):
        grp = descriptor['groups'][idx]
        # Verify group members must (exist in inputs, show up
        # once, only belong to single group)
        msg_template = " GroupError: \"{0}\" member \"{1}\" does not exist"
        errors += [msg_template.format(grp["id"], member)
                   for member in grp["members"] if member not in inIds]

        msg_template = " GroupError: \"{0}\" member \"{1}\" appears twice"
        errors += [msg_template.format(grp["id"], member)
                   for member in set(grp["members"])
                   if grp["members"].count(member) > 1]

        for jdx, grp2 in enumerate(descriptor["groups"]):
            msg_template = (" GroupError: \"{0}\" and \"{1}\" both contain"
                            " member \"{2}\"")
            errors += [msg_template.format(grp["id"], grp2["id"], m1)
                       for m1 in grp["members"]
                       for m2 in grp2["members"]
                       if m1 == m2 and idx > jdx]

        # Verify mutually exclusive groups cannot have required members
        # nor requiring members
        if "mutually-exclusive" in grp.keys():
            msg_template = (" GroupError: \"{0}\" is mutually-exclusive"
                            " and cannot have required members, "
                            "such as \"{1}\"")
            errors += [msg_template.format(grp["id"], member)
                       for member in set(grp["members"])
                       if not inById(member)["optional"]]

            msg_template = (" GroupError: \"{0}\" is mutually-exclusive"
                            " and cannot have members require one another,"
                            " such as \"{1}\" and \"{2}\"")
            for member in set(grp["members"]):
                if "requires-inputs" in inById(member).keys():
                    errors += [msg_template.format(grp["id"], member, req)
                               for req in inById(member)["requires-inputs"]
                               if req in set(grp["members"])]

        # Verify one-is-required groups should never have required members
        if "one-is-required" in grp.keys():
            msg_template = (" GroupError: \"{0}\" is a one-is-required"
                            " group and contains a required member, \"{1}\"")
            errors += [msg_template.format(grp["id"], member)
                       for member in set(grp["members"])
                       if member in inIds and not inById(member)["optional"]]

        # Verify one-is-required groups should never have required members
        if "all-or-none" in grp.keys():
            msg_template = (" GroupError: \"{0}\" is an all-or-none group"
                            " and cannot be paired with one-is-required"
                            " or mutually-exclusive groups")
            if ("one-is-required" in grp.keys()
               or "mutually-exclusive" in grp.keys()):
                    errors += [msg_template.format(grp["id"])]

            msg_template = (" GroupError: \"{0}\" is an all-or-none"
                            " group and contains a required member, \"{1}\"")
            errors += [msg_template.format(grp["id"], member)
                       for member in set(grp["members"])
                       if member in inIds and not inById(member)["optional"]]

    # Verify tests
    if "tests" in descriptor.keys():
        tests_names = []
        for test in descriptor["tests"]:

            tests_names.append(test["name"])
            if "output-files" in test["assertions"].keys():
                test_output_ids = safeGet(test["assertions"],
                                          "output-files", "id")

                # Verify if output reference ids are valid
                msg_template = ("TestError: \"{0}\" output id"
                                " not found, in test \"{1}\"")
                errors += [msg_template.format(output_id, test["name"])
                           for output_id in test_output_ids
                           if (output_id not in outIds)]

                # Verify that we do not have multiple output
                # references refering to the same id
                msg_template = ("TestError: \"{0}\" output id"
                                " cannot appear more than once within"
                                " same test, in test \"{1}\"")
                errors += [msg_template.format(output_id, test["name"])
                           for output_id in set(test_output_ids)
                           if (test_output_ids.count(output_id) > 1)]

        # Verify that all the defined tests have unique names
        msg_template = "TestError: \"{0}\" test name is non-unique"
        errors += [msg_template.format(test_name)
                   for test_name in set(tests_names)
                   if (tests_names.count(test_name) > 1)]

    errors = None if errors == [] else errors
    if errors is None:
        if kwargs.get("verbose"):
            print("Boutiques validation OK")
        return descriptor
    else:
        raise DescriptorValidationError("\n".join(errors))



import json
import tempfile
import os.path as op


def pytest_addoption(parser):
    parser.addoption("--descriptor", action="append", default=[])


def fetch_tests(descriptor_filename):

    with open(descriptor_filename) as fhandle:
        descriptor = json.loads(fhandle.read())

    tests = []

    # For each test present in the descriptor:
    for test in descriptor["tests"]:

        # We first extract the invocation and put it inside a temporary file.
        invocation_JSON = json.dumps(test["invocation"])
        temp_invocation_JSON = tempfile.NamedTemporaryFile(suffix=".json",
                                                           delete=False)
        temp_invocation_JSON.write(invocation_JSON.encode())
        temp_invocation_JSON.seek(0)

        # Now we setup the necessary elements for the testing function.
        tests.append([descriptor_filename, test, temp_invocation_JSON])

    return (descriptor["name"], tests)


# This function will be executed by pytest before the actual testing
def pytest_generate_tests(metafunc):
    descriptor_filename = metafunc.config.getoption('descriptor')[0]

    # Each element in 'tests' will hold the necessary informations
    # for a single test
    # Those informations are:
    #                         . The descriptor (common to all)
    #                         . The related JSON data, describing the test
    #                         (more convenient, no need to extract
    #                          again from descriptor)
    #                         .The invocation file needed for the test
    descriptor_name, tests = fetch_tests(descriptor_filename)

    # Generate the test ids for each of the test cases.
    # An id is created by concatenaning the name of the descriptor
    # with the name of the test case.
    names = ["{0}_{1}".format(op.basename(descriptor_filename),
             params[1]["name"].replace(' ', '-')) for params in tests]

    metafunc.parametrize("descriptor, test, invocation", tests, ids=names)



import sys
from setuptools import setup
import sys

VERSION = "0.5.9"
DEPS = [
         "simplejson",
         "requests",
         "pytest",
       ]

# jsonschema 2.6 doesn't support Python 2.6
# and unittest2 is a backport of Python 2.7's 'unittest to 2.6
if sys.version_info < (2, 7):
    DEPS.append("unittest2")
    DEPS.append("jsonschema==2.5.1")
else:
    DEPS.append("jsonschema")

setup(name="boutiques",
      version=VERSION,
      description="Schema for describing bash command-line tools",
      url="http://github.com/boutiques/boutiques",
      author="Tristan Glatard, Gregory Kiar",
      author_email="tristan.glatard@concordia.ca, gkiar07@gmail.com",
      classifiers=[
                "Programming Language :: Python",
                "Programming Language :: Python :: 2",
                "Programming Language :: Python :: 3",
                "Programming Language :: Python :: 2.6",
                "Programming Language :: Python :: 2.7",
                "Programming Language :: Python :: 3.5",
                "Programming Language :: Python :: 3.6",
                "Programming Language :: Python :: 3.7",
                "Programming Language :: Python :: Implementation :: PyPy",
                "License :: OSI Approved :: MIT License",
                "Topic :: Software Development :: Libraries :: Python Modules",
                "Operating System :: OS Independent",
                "Natural Language :: English"
                  ],
      license="MIT",
      packages=["boutiques"],
      include_package_data=True,
      test_suite="pytest",
      tests_require=["pytest"],
      setup_requires=DEPS,
      install_requires=DEPS,
      entry_points=  {
        "console_scripts": [
            "bosh=boutiques.bosh:bosh",
        ]
      },
      zip_safe=False)
